﻿using System;
using System.Linq;
using System.Text;
using NUnit.Framework;
using System.Net;
using System.Reflection;
using io = System.IO;

namespace CometDocs.Tests
{
    /// <summary>
    /// /!\ WARNING /!\
    /// WHILE THIS TEST SUITE HAS BEEN DESIGNED TO WORK SAFELY ON A SUB-FOLDER ONLY,
    /// YOU SHOULD USE IT ONLY WITH A COMETDOCS STORE DEDICATED TO TESTING.
    /// AVOID USING A STORE YOU USE FOR OTHER PURPOSES LIKE STORING YOUR DOCUMENTS.
    /// /!\ USE AT YOUR OWN RISK. /!\
    /// </summary>
    [TestFixture]
    public class ClientTests
    {
        private string UserEMail = null;
        private string UserPassword = null;
        private string UserAPIKey = null;
        private string TestFolderName = null;
        private FolderInfo TestFolderInfo = null;

        private static readonly Encoding encoding = Encoding.UTF8;

        private IClient client = null;
        private AuthenticationToken authToken = null;

        /// <summary>
        /// Read the testing credentials from "credentials.txt" file.
        /// </summary>
        private void ReadCredentials()
        {
            if (!io.File.Exists("Tests/credentials.txt"))
            {
                throw new Exception("The file 'Tests/credentials.txt' has not been found!\n" +
                                    "You must provide some credentials to use the CometDocs API.\n" +
                                    "Copy or rename the 'credentials.sample.txt' file and fill in your email, CometDocs password, CometDocs API key and testing folder name.");
            }

            string credentials = io.File.ReadAllText("Tests/credentials.txt");

            string lineEnding = credentials.Contains("\r\n") ? "\r\n" : "\n";

            string[] tokens = credentials.Split(new[] { lineEnding }, StringSplitOptions.RemoveEmptyEntries);

            if (tokens.Length != 4)
            {
                throw new Exception("You must provide exactly, and in the same order, your: email, password, API key and the tests folder!");
            }

            UserEMail = tokens[0];
            UserPassword = tokens[1];
            UserAPIKey = tokens[2];
            TestFolderName = tokens[3];
        }

        [TestFixtureSetUp]
        public void TestFixtureSetUp()
        {
            ReadCredentials();

            client = ClientFactory.GetClient();

            CanAuthenticate();

            Folder root = client.GetFolder(authToken);

            TestFolderInfo = root.Folders.SingleOrDefault(f => f.Name == TestFolderName);

            if (TestFolderInfo == null)
            {
                throw new Exception(string.Format("Unable to find tests folder '{0}'!\nPlease create it first.", TestFolderName));
            }
        }

        /// <summary>
        /// Ensure that the remote folder for testing is clean.
        /// </summary>
        private void AssertClean()
        {
            Folder testFolder = client.GetFolder(authToken, TestFolderInfo);

            foreach (FileInfo file in testFolder.Files)
            {
                client.DeleteFile(authToken, file);
            }

            foreach (FolderInfo folder in testFolder.Folders)
            {
                client.DeleteFolder(authToken, folder);
            }

            testFolder = client.GetFolder(authToken, TestFolderInfo);

            Assert.AreEqual(0, testFolder.Files.Length);
            Assert.AreEqual(0, testFolder.Folders.Length);
        }

        /// <summary>
        /// Test the "authenticate" method.
        /// It is invoked before each tests batch.
        /// </summary>
        public void CanAuthenticate()
        {
            authToken = client.Authenticate(UserEMail, UserPassword, UserAPIKey);

            Assert.That(!string.IsNullOrWhiteSpace(authToken.Value));
        }

        /// <summary>
        /// Ensure that a text file can be converted to PDF.
        /// </summary>
        [Test]
        public void CanConvertAFile()
        {
            AssertClean();

            File file = new File("a.txt")
            {
                Content = encoding.GetBytes("ABC")
            };

            FileInfo fileInfo = client.UploadFile(authToken, file, TestFolderInfo);

            Folder testFolder = client.GetFolder(authToken, TestFolderInfo);

            Assert.AreEqual(1, testFolder.Files.Length);
            Assert.AreEqual(fileInfo.ID, testFolder.Files[0].ID);
            Assert.AreEqual(fileInfo.Name, testFolder.Files[0].Name);
            Assert.AreEqual(fileInfo.Extension, testFolder.Files[0].Extension);

            ConversionType convType = new ConversionType("", "PDF");

            FileInfo pdfInfo = client.ConvertFile(authToken, fileInfo, convType);

            // Assert.That(pdfInfo.Extension.Equals("pdf", System.StringComparison.InvariantCultureIgnoreCase));

            testFolder = client.GetFolder(authToken, TestFolderInfo);

            Assert.AreEqual(1, testFolder.Files.Length);

            Assert.AreEqual(fileInfo.ID, testFolder.Files[0].ID);
            Assert.AreEqual(fileInfo.Name, testFolder.Files[0].Name);
            Assert.AreEqual(fileInfo.Extension, testFolder.Files[0].Extension);

            Conversion[] conversions = client.GetConversions(authToken, fileInfo);

            Assert.AreEqual(1, conversions.Length);

            Assert.AreEqual(pdfInfo.ID, conversions[0].ID);
            Assert.AreEqual(fileInfo.ID, conversions[0].ParentID);
            Assert.AreEqual(convType, conversions[0].Type);
            // Assert.AreEqual(pdfInfo.Extension, conversions[0].Extension);
        }

        /// <summary>
        /// Ensure that we can correctly represent files.
        /// </summary>
        [Test]
        public void CanCreateFileObject()
        {
            File file1 = new File("File1");

            Assert.AreEqual("File1", file1.Name);
            Assert.AreEqual(null, file1.Extension);

            File file2 = new File("File1.txt");

            Assert.AreEqual("File1", file2.Name);
            Assert.AreEqual("txt", file2.Extension);

            File file3 = new File("File2.txt.xml");

            Assert.AreEqual("File2.txt", file3.Name);
            Assert.AreEqual("xml", file3.Extension);
        }

        /// <summary>
        /// Ensure that we can create a new sub-folder in the store.
        /// </summary>
        [Test]
        public void CanCreateAFolder()
        {
            AssertClean();

            Folder testFolder = client.GetFolder(authToken, TestFolderInfo);

            String name = "__testFolder";

            FolderInfo newFolder = client.CreateFolder(authToken, testFolder, name);

            Assert.AreEqual(name, newFolder.Name);

            testFolder = client.GetFolder(authToken, TestFolderInfo);

            Folder match = testFolder.Folders.SingleOrDefault(f => f.ID == newFolder.ID);

            Assert.NotNull(match);

            Assert.AreEqual(newFolder.Name, match.Name);
        }

        /// <summary>
        /// Ensure we can create and delete a file in the store.
        /// </summary>
        [Test]
        public void CanDeleteAFile()
        {
            AssertClean();

            File file = new File("Test.txt");

            FileInfo info = client.UploadFile(authToken, file, TestFolderInfo);

            Folder testFolder = client.GetFolder(authToken, TestFolderInfo);

            Assert.AreEqual(1, testFolder.Files.Length);
            Assert.AreEqual(file.Name, testFolder.Files[0].Name);
            Assert.AreEqual(file.Extension, testFolder.Files[0].Extension);

            client.DeleteFile(authToken, info);

            testFolder = client.GetFolder(authToken, TestFolderInfo);

            Assert.AreEqual(0, testFolder.Files.Length);
        }

        /// <summary>
        /// Ensure we can get the whole content of a folder in the store.
        /// </summary>
        [Test]
        public void CanGetAFolder()
        {
            AssertClean();

            File file1 = new File("a.txt")
            {
                Content = encoding.GetBytes("ABC")
            };

            File file2 = new File("b.txt")
            {
                Content = encoding.GetBytes("BCD")
            };

            File file3 = new File("c.txt")
            {
                Content = encoding.GetBytes("CEF")
            };

            client.UploadFile(authToken, file1, TestFolderInfo);
            client.UploadFile(authToken, file2, TestFolderInfo);
            client.UploadFile(authToken, file3, TestFolderInfo);

            Folder testFolder = client.GetFolder(authToken, TestFolderInfo);

            Assert.NotNull(testFolder.Folders);

            Assert.NotNull(testFolder.Files);
            Assert.AreEqual(3, testFolder.Files.Length);

            Assert.AreEqual(file1.Name, testFolder.Files[0].Name);
            Assert.AreEqual(file1.Extension, testFolder.Files[0].Extension);

            Assert.AreEqual(file2.Name, testFolder.Files[1].Name);
            Assert.AreEqual(file2.Extension, testFolder.Files[1].Extension);

            Assert.AreEqual(file3.Name, testFolder.Files[2].Name);
            Assert.AreEqual(file3.Extension, testFolder.Files[2].Extension);
        }

        /// <summary>
        /// Ensure we can get the list of all the files categories.
        /// </summary>
        [Test]
        public void CanGetCategories()
        {
            Category[] categories = client.GetCategories();

            Assert.GreaterOrEqual(categories.Length, 10);
            Assert.That(categories.Any(cat => cat.Name == "Art"));
            Assert.That(categories.Any(cat => cat.Name == "Business"));
            Assert.That(!categories.Any(cat => cat.Name == "Unicorns"));
        }

        /// <summary>
        /// Ensure we can get the list of all the available conversions.
        /// </summary>
        [Test]
        public void CanGetConversionTypes()
        {
            ConversionType[] types = client.GetConversionTypes();

            Assert.GreaterOrEqual(types.Length, 10);
            Assert.That(types.Any(t => t.ToString() == "PDF2DOC"));
            Assert.That(types.Any(t => t.ToString() == "PDF2XLS"));
            Assert.That(!types.Any(t => t.ToString() == "PDF2XYZ"));
        }

        /// <summary>
        /// Ensure we can get the list of all the CometDocs API entry-points, and that they match all the methods implemented in the Client class.
        /// </summary>
        [Test]
        public void CanGetMethods()
        {
            string[] methods = client.GetMethods();

            string[] methodsName = methods.Select(m => m.Split('(')[0]).OrderBy(m => m).ToArray();
            string[] clientMethods = typeof(Client).GetMethods(BindingFlags.Public | BindingFlags.Instance | BindingFlags.DeclaredOnly).Select(m => m.Name).Distinct().OrderBy(m => m).ToArray();

            Assert.That(clientMethods.SequenceEqual(methodsName, StringComparer.InvariantCultureIgnoreCase));
        }

        /// <summary>
        /// Ensure we can get the list of CometDocs stores public files.
        /// </summary>
        [Test]
        public void CanGetPublicFiles()
        {
            FileInfo[] files = client.GetPublicFiles(authToken);

            Assert.GreaterOrEqual(files.Length, 100);
            Assert.That(files.Any(file => file.Extension == "pdf"));
            Assert.That(files.Any(file => file.Extension == "jpg"));
        }

        /// <summary>
        /// Ensure we can get the list of all the shared file.
        /// As of 2013-06-26 only checks there is no error and the list is empty, because there is no entry-point for sharing a file.
        /// </summary>
        [Test]
        public void CanGetSharedFiles()
        {
            FileInfo[] files = client.GetSharedFiles(authToken);

            Assert.AreEqual(0, files.Length);
        }

        /// <summary>
        /// Ensure we can invalidate an authentication token.
        /// </summary>
        [Test]
        public void CanInvalidateToken()
        {
            AssertClean();

            client.GetFolder(authToken);

            client.InvalidateToken(authToken);

            Assert.Throws<InvalidTokenException>(() => client.GetFolder(authToken));

            // Ensure we're authenticated again for next tests
            CanAuthenticate();
        }

        /// <summary>
        /// Ensure we can send a file by mail.
        /// </summary>
        [Test]
        public void CanSendAFile()
        {
            AssertClean();

            File file = new File("a.txt")
            {
                Content = encoding.GetBytes("ABC")
            };

            FileInfo fileInfo = client.UploadFile(authToken, file, TestFolderInfo);

            client.SendFile(authToken, fileInfo, new[] { UserEMail }, "some.user@company.com", "Hello!");

            // Assert you've received the email :)
        }

        /// <summary>
        /// Ensure we can upload a file from memory to the store and get it back.
        /// </summary>
        [Test]
        public void CanUploadAndDownloadAFile()
        {
            AssertClean();

            string content = @"The peanut, or groundnut (Arachis hypogaea), is a species in the legume or ""bean"" family (Fabaceae). The peanut was probably first domesticated and cultivated in the valleys of Paraguay.[1] It is an annual herbaceous plant growing 30 to 50 cm (1.0 to 1.6 ft) tall. The leaves are opposite, pinnate with four leaflets (two opposite pairs; no terminal leaflet), each leaflet is 1 to 7 cm (⅜ to 2¾ in) long and 1 to 3 cm (⅜ to 1 inch) broad.
The flowers are a typical peaflower in shape, 2 to 4 cm (0.8 to 1.6 in) (¾ to 1½ in) across, yellow with reddish veining. Hypogaea means ""under the earth""; after pollination, the flower stalk elongates causing it to bend until the ovary touches the ground. Continued stalk growth then pushes the ovary underground where the mature fruit develops into a legume pod, the peanut – a classical example of geocarpy. Pods are 3 to 7 cm (1.2 to 2.8 in) long, containing 1 to 4 seeds.[2]
Peanuts are known by many other local names such as earthnuts, ground nuts, goober peas, monkey nuts, pygmy nuts and pig nuts.[3] Despite its name and appearance, the peanut is not a nut, but rather a legume.";
            File inputFile = new File("Peanuts.txt")
            {
                Content = encoding.GetBytes(content)
            };

            FileInfo info = client.UploadFile(authToken, inputFile, TestFolderInfo);
            Assert.AreEqual(inputFile.Content.Length, info.Size);

            File outputFile = client.DownloadFile(authToken, info);

            Assert.That(inputFile.Content.SequenceEqual(outputFile.Content));

            string outputContent = encoding.GetString(outputFile.Content);

            Assert.AreEqual(content, outputContent);
        }

        /// <summary>
        /// Ensure we can upload an internet public file to our store.
        /// </summary>
        [Test]
        public void CanUploadAFileFromUrl()
        {
            AssertClean();

            string fileUrl = "http://www.imf.org/external/np/res/commod/Table4.pdf";

            client.UploadFileFromUrl(authToken, fileUrl, TestFolderInfo);

            Folder testFolder = client.GetFolder(authToken, TestFolderInfo);

            FileInfo info = testFolder.Files[0];

            Assert.AreEqual(1, testFolder.Files.Length);
            Assert.AreEqual("Table4", info.Name);
            Assert.AreEqual("pdf", info.Extension);

            File outputFile = client.DownloadFile(authToken, info);

            WebClient wc = new WebClient();
            byte[] fileContent = wc.DownloadData(fileUrl);

            Assert.That(outputFile.Content.SequenceEqual(fileContent));
        }
    }
}
